import hid

from usbfan.protocol import Program


class Device:
    def __init__(self, vendor_id=0x0c45, product_id=0x7160):
        self.device = hid.device()
        self.device.open(vendor_id, product_id)

    def program(self, program: Program):
        if not isinstance(program, Program):
            raise ValueError("need a Program to program the device")

        for message in program:
            self.device.write(message)
            ack = bytes(self.device.read(3))
            if ack != b'\x24\x80':
                print("Got an error on upload")


if __name__ == "__main__":
    from usbfan.message import TextMessage

    f = Program((TextMessage("Hello, World!"),
                 TextMessage("How is everyone going?"),))
    d = Device()
    d.program(f)



from os.path import basename, join

from PIL import Image, ImageDraw, ImageFont

from .protocol import Colour, Column, Message


class TextMessage(Message):
    def __init__(self, text: str):
        # Open the font, create a blank image and an ImageDraw object
        fnt = ImageFont.truetype(
            join(basename(__file__), 'fonts', 'Hack-Regular.ttf'))
        img = Image.new('RGB', (Message.MAX_COLUMNS, Column.PIXELS))
        d = ImageDraw.Draw(img)

        # Write the text into the image
        d.text((0, -1), text, font=fnt)

        # Transpose the image
        img = img.transpose(Image.TRANSPOSE)

        # Convert the image into one channel
        img_data = [True if p >= 128 else False for p in
                    img.convert('L').getdata(0)]

        # Convert the image into its columns
        columns = [Column(img_data[i:i+Column.PIXELS], Colour.white) for i in
                   range(0, len(img_data), Column.PIXELS)]

        super().__init__(columns)



from enum import IntEnum
from math import floor, log2
from typing import Iterator, Optional, Sequence, Union


class Colour(IntEnum):
    # primary
    red = 0b00100000_00000000
    blue = 0b01000000_00000000
    green = 0b10000000_00000000

    # mixes
    magenta = 0b01100000_00000000
    yellow = 0b10100000_00000000
    cyan = 0b11000000_00000000

    # everything
    white = 0b11100000_00000000


class Column:
    PIXELS = 11

    def __init__(self, pixels: Optional[Sequence[bool]]=None,
                 colour: Union[int, Colour]=Colour.red):
        if pixels:
            try:
                if not all(isinstance(p, bool) for p in pixels):
                    raise ValueError("pixels must be an iterable of bool's")
                elif len(pixels) != Column.PIXELS:
                    raise ValueError(f"pixels must be an iterable of "
                                     f"{Column.PIXELS} bool's")
            except TypeError:
                raise ValueError("pixels must be an iterable of bool's")
            self.pixels = list(pixels)
        else:
            self.pixels = [False] * Column.PIXELS

        if not isinstance(colour, (int, Colour)):
            raise ValueError('colour must be an int or Colour')
        if colour & Colour.white != colour:
            raise ValueError('bad colour value')
        self.colour = colour

    def __bytes__(self) -> bytes:
        b = 0x0000
        for i in range(Column.PIXELS):
            if self.pixels[i]:
                b |= 1 << i

        b |= self.colour
        return bytes([(b & 0xFF00) >> 8, b & 0x00FF])

    def __len__(self) -> int:
        return Column.PIXELS

    def __getitem__(self, key: int) -> bool:
        return self.pixels[key]

    def __setitem__(self, key: int, value: bool) -> None:
        self.pixels[key] = value

    def __iter__(self) -> Iterator[bool]:
        return iter(self.pixels)


class Message:
    MAX_COLUMNS = 144
    COLUMN_GROUPS = 8

    def __init__(self, columns: Optional[Sequence[Column]]=None):
        if columns:
            try:
                if not all(isinstance(x, Column) for x in columns):
                    raise ValueError("Message must be an iterable of Columns")
                elif len(columns) > Message.MAX_COLUMNS:
                    raise ValueError(f"Message must be an iterable of no more "
                                     f"than {Message.MAX_COLUMNS} Columns")
                elif len(columns) % Message.COLUMN_GROUPS != 0:
                    raise ValueError(f"Message must be an iterable with a "
                                     f"multiple of {Message.COLUMN_GROUPS} "
                                     f"columns")
            except TypeError:
                raise ValueError("Message must be an iterable of Columns")

            self.columns = list(columns)
        else:
            self.columns = [Column() for _ in range(Message.MAX_COLUMNS)]

    def __len__(self) -> int:
        return len(self.columns)

    def __getitem__(self, key: int) -> Column:
        return self.columns[key]

    def __setitem__(self, key: int, value: Column) -> None:
        self.columns[key] = value

    def __bytes__(self) -> bytes:
        # start with the program length
        # note: we always have two NUL pixels at the start and end
        program_data = bytes([len(self.columns) + 2])

        # followed by 3 + 2 NUL's (possible that some of these are the style?)
        program_data += b'\0\0\0\0\0'

        # concatenate the bytes of each column - columns are put in backwards
        program_data += b''.join(bytes(x) for x in reversed(self.columns))

        # finally, two more NUL's
        program_data += b'\0\0'

        # apply subtraction
        program_data = bytes((Program.SUBTRACTION - x) & 0xFF
                             for x in program_data)

        # build the iterable
        return program_data


class MessageIterator:
    FIRST_HEADER = b'\x00\x40\x40'
    HEADER = b'\x00\x40\x23'

    def __init__(self, data: bytes):
        self.data = data
        self._pos = 0

    def __iter__(self):
        return self

    def __next__(self) -> bytes:
        if self._pos + 5 > len(self.data):
            raise StopIteration
        elif self._pos > 0:
            message = Program.checksum(
                MessageIterator.HEADER + self.data[self._pos:self._pos + 5])
        else:
            message = Program.checksum(
                MessageIterator.FIRST_HEADER +
                self.data[self._pos:self._pos+5])
        self._pos += 5
        return message


class Program:
    SUBTRACTION = 0xA4
    MAX_MESSAGES = 7

    def __init__(self, messages: Sequence[Message]):
        if len(messages) > Program.MAX_MESSAGES:
            raise ValueError("Too many messages provided")
        self.messages = list(messages)

    @staticmethod
    def checksum(message: bytes) -> bytes:
        if not isinstance(message, bytes):
            raise ValueError("Can only checksum bytes type")
        if len(message) != 8:
            raise ValueError("Incorrect message length for checksumming")

        return message + bytes([sum(message[1:8]) & 0xFF])

    def __len__(self) -> int:
        return len(self.messages)

    def __getitem__(self, key: int) -> Message:
        return self.messages[key]

    def __setitem__(self, key: int, value: Message) -> None:
        self.messages[key] = value

    def __iter__(self):
        # first get all the program data, plus 10 NUL's to mark the end
        messages = b''.join(bytes(p) for p in self.messages) + \
                   bytes([Program.SUBTRACTION]) * 10

        # data starts with the following algorithm:
        #     max(floor(log_2(data_length) - 6), 1)
        # the program has this as a series of if's, but this is Python
        data = bytes([floor(log2(max(len(messages) >> 6, 2))),
                      len(messages) & 0x00FF,
                      (len(messages) & 0xFF00) >> 8])

        # three NUL's follow, however the 3rd NUL falls into the 2nd message,
        # which is sent under the SUBTRACTION, therefore it is 0xA4
        data += b'\0\0' + bytes([Program.SUBTRACTION])

        # next, the number of programs is sent
        data += bytes([0x24 - len(self.messages)])

        # get the data for each of the programs
        data += messages

        return MessageIterator(data)



from setuptools import setup


install_requires = [pkg.strip() for pkg in open('requirements.txt').readlines()]

setup(name='microwave-usbfan',
      version='1.0',
      description="Implementation of the Jaycar RGB USB LED fan (GH1031) protocol",
      long_description=open('README.md').read(),
      url='https://github.com/fergofrog/microwave_usb_fan',
      author='Fergus Symon',
      author_email='45412+fergofrog@users.noreply.github.com',
      license='MIT',
      packages=['usbfan'],
      install_requires=install_requires,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: MacOS',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX :: BSD',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Communications',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      )



from unittest import TestCase

from usbfan import Colour, Column, Message, Program


class TestColumn(TestCase):
    def test_blank(self):
        c = Column()

        self.assertEqual(len(c), 11, 'wrong number of pixels')
        for pixel in c:
            self.assertEqual(pixel, False, "pixel isn't 'False'")
        self.assertEqual(bytes(c), b'\x20\x00', 'column data is wrong')

    def test_mutation(self):
        c = Column()

        self.assertEqual(c[3], False, "4th pixel isn't 'False'")
        c[3] = True
        self.assertEqual(c[3], True, "4th pixel isn't 'True' after mutation")
        self.assertEqual(bytes(c), b'\x20\x08', 'column data is wrong')

    def test_custom(self):
        pixels = (True, False, True, False, True, False,
                  True, False, True, False, True)
        c = Column(pixels, Colour.cyan)

        self.assertSequenceEqual(c, pixels, 'column data taken from iter wrong')
        self.assertEqual(bytes(c), b'\xc5\x55', 'column data is wrong')

    def test_bad_data(self):
        with self.assertRaises(ValueError):
            Column('abcdefghijk')
        with self.assertRaises(ValueError):
            Column(colour=0x20)
        with self.assertRaises(ValueError):
            Column((True, False, True, False))


class TestMessage(TestCase):
    def test_blank(self):
        p = Message()

        self.assertEqual(len(p), 144)
        for column in p:
            self.assertIsInstance(column, Column,
                                  'Column not contained in Message iter')
        self.assertEqual(bytes(p),
                         b'\x12\xa4\xa4\xa4\xa4\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa4'
                         b'\x84\xa4\x84\xa4\x84\xa4\x84\xa4\xa4\xa4',
                         "empty program does't match reference data")

    def test_small(self):
        columns = [Column([True] + [False] * 10, Colour.red)]
        for _ in range(7):
            columns.append(Column([False] * 11, Colour.red))

        p = Message(columns)

        self.assertEqual(len(p), 8)
        self.assertEqual(bytes(p),
                         b'\x9a\xa4\xa4\xa4\xa4\xa4\x84\xa4\x84\xa4\x84\xa4\x84'
                         b'\xa4\x84\xa4\x84\xa4\x84\xa4\x84\xa3\xa4\xa4',
                         "small red dot doesn't match reference data")


class TestProgram(TestCase):
    def test_checksum(self):
        self.assertEqual(Program.checksum(b'\x00\x40\x40\x01\xA2\x00\x00\x00'),
                         b'\x00\x40\x40\x01\xA2\x00\x00\x00\x23',
                         "checksum doesn't match reference data")
        self.assertEqual(Program.checksum(b'\x00\x40\x23\x80\x83\x80\x83\x80'),
                         b'\x00\x40\x23\x80\x83\x80\x83\x80\xE9',
                         "checksum doesn't match reference data")
        with self.assertRaises(ValueError):
            Program.checksum('abcdefgh')
        with self.assertRaises(ValueError):
            Program.checksum(b'aoeu')

    def test_small(self):
        columns = [Column([True] + [False] * 10, Colour.red)]
        for _ in range(7):
            columns.append(Column([False] * 11, Colour.red))
        f = Program((Message(columns),))

        self.assertEqual(len(f), 1)
        self.assertIsInstance(f[0], Message)

        reference_message = (
            b'\x00\x40\x40\x01\x22\x00\x00\x00\xA3',
            b'\x00\x40\x23\xA4\x23\x9A\xA4\xA4\x0C',
            b'\x00\x40\x23\xA4\xA4\xA4\x84\xA4\x77',
            b'\x00\x40\x23\x84\xA4\x84\xA4\x84\x37',
            b'\x00\x40\x23\xA4\x84\xA4\x84\xA4\x57',
            b'\x00\x40\x23\x84\xA4\x84\xA3\xA4\x56',
            b'\x00\x40\x23\xA4\xA4\xA4\xA4\xA4\x97',
        )

        for data, reference in zip(f, reference_message):
            self.assertEqual(data, reference)

    def test_large(self):
        programs = list()
        for colour in (Colour.red, Colour.green, Colour.blue, Colour.yellow,
                       Colour.magenta, Colour.cyan, Colour.white):
            columns = [Column([True] + [False] * 10, colour)]
            for _ in range(14):
                columns.append(Column([False] * 11, colour))
            columns.append(Column([False] * 10 + [True], colour))
            programs.append(Message(columns))
        f = Program(programs)

        self.assertEqual(len(f), 7)
        self.assertIsInstance(f[6], Message)

        reference_message = (
            b'\x00\x40\x40\x02\x22\x01\x00\x00\xA5',
            b'\x00\x40\x23\xA4\x1D\x92\xA4\xA4\xFE',
            b'\x00\x40\x23\xA4\xA4\xA4\x80\xA4\x73',
            b'\x00\x40\x23\x84\xA4\x84\xA4\x84\x37',
            b'\x00\x40\x23\xA4\x84\xA4\x84\xA4\x57',
            b'\x00\x40\x23\x84\xA4\x84\xA4\x84\x37',
            b'\x00\x40\x23\xA4\x84\xA4\x84\xA4\x57',
            b'\x00\x40\x23\x84\xA4\x84\xA4\x84\x37',
            b'\x00\x40\x23\xA4\x84\xA4\x84\xA3\x56',
            b'\x00\x40\x23\xA4\xA4\x92\xA4\xA4\x85',
            b'\x00\x40\x23\xA4\xA4\xA4\x20\xA4\x13',
            b'\x00\x40\x23\x24\xA4\x24\xA4\x24\x17',
            b'\x00\x40\x23\xA4\x24\xA4\x24\xA4\x97',
            b'\x00\x40\x23\x24\xA4\x24\xA4\x24\x17',
            b'\x00\x40\x23\xA4\x24\xA4\x24\xA4\x97',
            b'\x00\x40\x23\x24\xA4\x24\xA4\x24\x17',
            b'\x00\x40\x23\xA4\x24\xA4\x24\xA3\x96',
            b'\x00\x40\x23\xA4\xA4\x92\xA4\xA4\x85',
            b'\x00\x40\x23\xA4\xA4\xA4\x60\xA4\x53',
            b'\x00\x40\x23\x64\xA4\x64\xA4\x64\xD7',
            b'\x00\x40\x23\xA4\x64\xA4\x64\xA4\x17',
            b'\x00\x40\x23\x64\xA4\x64\xA4\x64\xD7',
            b'\x00\x40\x23\xA4\x64\xA4\x64\xA4\x17',
            b'\x00\x40\x23\x64\xA4\x64\xA4\x64\xD7',
            b'\x00\x40\x23\xA4\x64\xA4\x64\xA3\x16',
            b'\x00\x40\x23\xA4\xA4\x92\xA4\xA4\x85',
            b'\x00\x40\x23\xA4\xA4\xA4\x00\xA4\xF3',
            b'\x00\x40\x23\x04\xA4\x04\xA4\x04\xB7',
            b'\x00\x40\x23\xA4\x04\xA4\x04\xA4\x57',
            b'\x00\x40\x23\x04\xA4\x04\xA4\x04\xB7',
            b'\x00\x40\x23\xA4\x04\xA4\x04\xA4\x57',
            b'\x00\x40\x23\x04\xA4\x04\xA4\x04\xB7',
            b'\x00\x40\x23\xA4\x04\xA4\x04\xA3\x56',
            b'\x00\x40\x23\xA4\xA4\x92\xA4\xA4\x85',
            b'\x00\x40\x23\xA4\xA4\xA4\x40\xA4\x33',
            b'\x00\x40\x23\x44\xA4\x44\xA4\x44\x77',
            b'\x00\x40\x23\xA4\x44\xA4\x44\xA4\xD7',
            b'\x00\x40\x23\x44\xA4\x44\xA4\x44\x77',
            b'\x00\x40\x23\xA4\x44\xA4\x44\xA4\xD7',
            b'\x00\x40\x23\x44\xA4\x44\xA4\x44\x77',
            b'\x00\x40\x23\xA4\x44\xA4\x44\xA3\xD6',
            b'\x00\x40\x23\xA4\xA4\x92\xA4\xA4\x85',
            b'\x00\x40\x23\xA4\xA4\xA4\xE0\xA4\xD3',
            b'\x00\x40\x23\xE4\xA4\xE4\xA4\xE4\x57',
            b'\x00\x40\x23\xA4\xE4\xA4\xE4\xA4\x17',
            b'\x00\x40\x23\xE4\xA4\xE4\xA4\xE4\x57',
            b'\x00\x40\x23\xA4\xE4\xA4\xE4\xA4\x17',
            b'\x00\x40\x23\xE4\xA4\xE4\xA4\xE4\x57',
            b'\x00\x40\x23\xA4\xE4\xA4\xE4\xA3\x16',
            b'\x00\x40\x23\xA4\xA4\x92\xA4\xA4\x85',
            b'\x00\x40\x23\xA4\xA4\xA4\xC0\xA4\xB3',
            b'\x00\x40\x23\xC4\xA4\xC4\xA4\xC4\xF7',
            b'\x00\x40\x23\xA4\xC4\xA4\xC4\xA4\xD7',
            b'\x00\x40\x23\xC4\xA4\xC4\xA4\xC4\xF7',
            b'\x00\x40\x23\xA4\xC4\xA4\xC4\xA4\xD7',
            b'\x00\x40\x23\xC4\xA4\xC4\xA4\xC4\xF7',
            b'\x00\x40\x23\xA4\xC4\xA4\xC4\xA3\xD6',
            b'\x00\x40\x23\xA4\xA4\xA4\xA4\xA4\x97',
            b'\x00\x40\x23\xA4\xA4\xA4\xA4\xA4\x97',
            b'\x00\x40\x23\x64\xA4\xA4\xA4\xA4\x57',
            b'\x00\x40\x23\xA4\xA4\xA4\xA4\xA4\x97'
        )

        for data, reference in zip(f, reference_message):
            self.assertEqual(data, reference)
